/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "OctreeAuxiliary.h"   /// own header is included last


using namespace hxa7241::graphics;




/// OctreeDimensions ///////////////////////////////////////////////////////////////////////////////


/// to fit within fp single precision
const dword OctreeDimensions::MAX_LEVEL = 23;


/// standard object services -----------------------------------------------------------------------
OctreeDimensions::OctreeDimensions
(
	const Vector3f& positionOfLowerCorner,
	const float     size,
	const dword     maxElementsPerCell,
	const dword     maxLevelCount
)
 :	positionOfLowerCorner_m( positionOfLowerCorner ),
	size_m                 ( size               >= 0.0f ? size : -size ),
	maxElementsPerCell_m   ( maxElementsPerCell >  0    ? maxElementsPerCell : 1 ),
	maxLevel_m             ( maxLevelCount      >  0    ? maxLevelCount - 1  : 0 )
{
	if( maxLevel_m > MAX_LEVEL )
	{
		maxLevel_m = MAX_LEVEL;
	}
}


OctreeDimensions::~OctreeDimensions()
{
}


OctreeDimensions::OctreeDimensions
(
	const OctreeDimensions& other
)
// :	positionOfLowerCorner_m( other.positionOfLowerCorner_m ),
//	size_m                 ( other.size_m ),
//	maxElementsPerCell_m   ( other.maxElementsPerCell_m ),
//	maxLevel_m             ( other.maxLevel_m )
{
	OctreeDimensions::operator=( other );
}


OctreeDimensions& OctreeDimensions::operator=
(
	const OctreeDimensions& other
)
{
	if( &other != this )
	{
		positionOfLowerCorner_m = other.positionOfLowerCorner_m;
		size_m                  = other.size_m;
		maxElementsPerCell_m    = other.maxElementsPerCell_m;
		maxLevel_m              = other.maxLevel_m;
	}

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
bool OctreeDimensions::isSubdivide
(
	const dword elementCount,
	const dword level
) const
{
	return (elementCount > maxElementsPerCell_m) & (level < maxLevel_m);
}








/// OctreeBound ////////////////////////////////////////////////////////////////////////////////////


/// standard object services -----------------------------------------------------------------------
OctreeBound::OctreeBound()
 :	positionOfLowerCorner_m( Vector3f::ZERO() ),
	positionOfUpperCorner_m( Vector3f::ONE() ),
	center_m               ( Vector3f::HALF() ),
	circumSphereRadius_m   ( Vector3f::HALF().length() )
{
}


OctreeBound::OctreeBound
(
	const Vector3f& positionOfLowerCorner,
	const float     size
)
 :	positionOfLowerCorner_m( positionOfLowerCorner ),
	positionOfUpperCorner_m( positionOfLowerCorner + Vector3f(size, size, size) ),
	center_m               ( (positionOfLowerCorner_m + positionOfUpperCorner_m) *= 0.5f ),
	circumSphereRadius_m   ( (Vector3f::HALF() * size).length() )
{
}


OctreeBound::OctreeBound
(
	const OctreeBound& parentCellBound,
	const dword        subCellIndex
)
{
	{
		const Vector3f* lowMidHigh[]  = { &(parentCellBound.positionOfLowerCorner_m),
		                                  &(parentCellBound.center_m),
		                                  &(parentCellBound.positionOfUpperCorner_m) };

		positionOfLowerCorner_m.setXYZ( lowMidHigh[ subCellIndex       & 1]->getX(),
		                                lowMidHigh[(subCellIndex >> 1) & 1]->getY(),
		                                lowMidHigh[(subCellIndex >> 2) & 1]->getZ() );
		positionOfUpperCorner_m.setXYZ( (lowMidHigh+1)[ subCellIndex       & 1]->getX(),
		                                (lowMidHigh+1)[(subCellIndex >> 1) & 1]->getY(),
		                                (lowMidHigh+1)[(subCellIndex >> 2) & 1]->getZ() );
	}

	((center_m = positionOfLowerCorner_m) += positionOfUpperCorner_m) *= 0.5f;
	circumSphereRadius_m = parentCellBound.circumSphereRadius_m * 0.5f;
}


OctreeBound::~OctreeBound()
{
}


OctreeBound::OctreeBound
(
	const OctreeBound& other
)
 :	positionOfLowerCorner_m( other.positionOfLowerCorner_m ),
	positionOfUpperCorner_m( other.positionOfUpperCorner_m ),
	center_m               ( other.center_m ),
	circumSphereRadius_m   ( other.circumSphereRadius_m )
{
}


OctreeBound& OctreeBound::operator=
(
	const OctreeBound& other
)
{
	if( &other != this )
	{
		positionOfLowerCorner_m = other.positionOfLowerCorner_m;
		positionOfUpperCorner_m = other.positionOfUpperCorner_m;
		center_m                = other.center_m;
		circumSphereRadius_m    = other.circumSphereRadius_m;
	}

	return *this;
}








/// OctreeData /////////////////////////////////////////////////////////////////////////////////////


/// standard object services -----------------------------------------------------------------------
OctreeData::OctreeData
(
	const OctreeDimensions& dimensions
)
 :	bound_m      ( dimensions.getPosition(), dimensions.getSize() ),
	level_m      ( 0 ),
	pDimensions_m( &dimensions )
{
}


OctreeData::OctreeData
(
	const OctreeData& parentCellData,
	const dword       subCellIndex
)
 :	bound_m      ( parentCellData.bound_m, subCellIndex ),
	level_m      ( parentCellData.level_m + 1 ),
	pDimensions_m( parentCellData.pDimensions_m )
{
}


OctreeData::OctreeData
(
	const OctreeData&       other,
	const OctreeDimensions& dimensions
)
 :	bound_m      ( other.bound_m ),
	level_m      ( other.level_m ),
	pDimensions_m( &dimensions )
{
}


OctreeData::~OctreeData()
{
}




/// commands ---------------------------------------------------------------------------------------
OctreeData& OctreeData::assign
(
	const OctreeData&       other,
	const OctreeDimensions& dimensions
)
{
	if( &other != this )
	{
		bound_m       = other.bound_m;
		level_m       = other.level_m;
		pDimensions_m = &dimensions;
	}

	return *this;
}
