/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "OctreeVisitorPhoton.h"
#include "Light.h"
#include "Vector3f.h"
#include "UnitVector3f.h"
#include "CieXyz.h"
#include "hxa7241math.h"
#include "Filters.h"
#include "Float01.h"
#include "Logger.h"

#include "PhotonMap.h"   /// own header is included last


using namespace perceptuum2::transport;




//#define ENABLE_TEST_CODE


/// statics
const dword PhotonMap::MAX_TREE_LEVELS          = 128;
const dword PhotonMap::MAX_TREE_CELL_ITEM_COUNT =  96;
const dword PhotonMap::ESTIMATION_PROPORTION    = 400;   /// DEBUG ///
const dword PhotonMap::ESTIMATION_SIZE          =  48;   /// DEBUG ///




/// standard object services -----------------------------------------------------------------------
PhotonMap::PhotonMap
(
	PhotonMapBuffers* pBuffers,
	const dword       capacity,
	const bool        isLogging
)
 :	maxOctreeLevels_m          ( MAX_TREE_LEVELS ),
	maxItemCountPerOctreeCell_m( MAX_TREE_CELL_ITEM_COUNT ),
	photons_m           (),
	minPosition_m       ( Vector3f::MAX() ),
	maxPosition_m       ( Vector3f::MIN() ),
	pOctree_m           (),
	octreeMaxDepth_m    (),
	octreeElementCount_m(),
	pBuffers_m          ( pBuffers ),
	totalEstimationCount_m( 0 ),
	totalVisitCount_m     ( 0 ),
	totalShrinkCount_m    ( 0 ),
	maxVisitCount_m       ( 0 ),
	maxShrinkCount_m      ( 0 ),
	isLogging_m           ( isLogging )
{
	photons_m.reserve( capacity >= 0 ? capacity : 0 );

	if( isLogging_m )
	{
		hxa7241::utility::Logger() << "PhotonMap -  constructed,  photons size: " << photons_m.size() << "  photons capacity: " << photons_m.capacity();
	}
}


PhotonMap::PhotonMap
(
	PhotonMapBuffers* pBuffers,
	const dword       capacity,
	const dword       maxTreeLevels,
	const dword       maxTreeCellItemCount,
	const bool        isLogging
)
 :	maxOctreeLevels_m          ( maxTreeLevels        > 0 ? maxTreeLevels        : 1 ),
	maxItemCountPerOctreeCell_m( maxTreeCellItemCount > 0 ? maxTreeCellItemCount : 1 ),
	photons_m           (),
	minPosition_m       ( Vector3f::MAX() ),
	maxPosition_m       ( Vector3f::MIN() ),
	pOctree_m           (),
	octreeMaxDepth_m    (),
	octreeElementCount_m(),
	pBuffers_m          ( pBuffers ),
	totalEstimationCount_m( 0 ),
	totalVisitCount_m     ( 0 ),
	totalShrinkCount_m    ( 0 ),
	maxVisitCount_m       ( 0 ),
	maxShrinkCount_m      ( 0 ),
	isLogging_m           ( isLogging )
{
	photons_m.reserve( capacity >= 0 ? capacity : 0 );

	if( isLogging_m )
	{
		hxa7241::utility::Logger() << "PhotonMap -  constructed,  photons size: " << photons_m.size() << "  photons capacity: " << photons_m.capacity();
	}
}


PhotonMap::~PhotonMap()
{
}




/// commands ---------------------------------------------------------------------------------------
void PhotonMap::switchToBuildingMode()
{
	/// check whether to init the photons store and stats, and destroy the octree
	if( pOctree_m.get() != 0 )
	{
		photons_m.resize( 0 );
		minPosition_m = Vector3f::MAX();
		maxPosition_m = Vector3f::MIN();

		pOctree_m.reset( 0 );
		octreeMaxDepth_m     = 0;
		octreeElementCount_m = 0;
	}
}


void PhotonMap::switchToIlluminatingMode()
{
	/// check whether to init the octree
	if( pOctree_m.get() == 0 )
	{
		/// reserve buffer
		if( pBuffers_m != 0 )
		{
			//const dword estimationSize = (photons_m.size() / ESTIMATION_PROPORTION) > 0 ? (photons_m.size() / ESTIMATION_PROPORTION) : 1;
			const dword estimationSize = ESTIMATION_SIZE;   /// DEBUG ///
			pBuffers_m->reserve( estimationSize );
		}

		/// get octree size from photons stats
		float octreeWidth = 0.0f;
		if( photons_m.size() > 0 )
		{
			octreeWidth = (maxPosition_m - minPosition_m).largest();
			/// add tolerance of a rough epsilon at this scale (1/2^20th * width)
			octreeWidth += (octreeWidth / float(0x00100000));
		}
		else
		{
			minPosition_m = Vector3f::ZERO();
			maxPosition_m = Vector3f::ZERO();
		}

		/// make octree
		pOctree_m.reset( new Octree<Photon>( minPosition_m, octreeWidth,
											 maxItemCountPerOctreeCell_m, maxOctreeLevels_m ) );

		/// put photons into octree
		OctreeVisitorPhoton octreeVisitor;
		for( dword i = photons_m.size();  i-- > 0; )
		{
			pOctree_m->insertItem( photons_m[i], octreeVisitor );
		}

		/// get octree info
		{
			dword byteSize;
			dword leafCount;
			dword elementCount;
			dword maxDepth;
			pOctree_m->getInfo( byteSize, leafCount, elementCount, maxDepth );

			octreeMaxDepth_m     = maxDepth;
			octreeElementCount_m = elementCount;

			if( isLogging_m )
			{
				hxa7241::utility::Logger() << "PhotonMap -  photons info: elementcount=" << photons_m.size() << " bytesize=" << (photons_m.size() * sizeof(Photon)) << "  octree info: byteSize=" << byteSize << " leafCount=" << leafCount << " elementCount=" << elementCount << " maxDepth=" << maxDepth;
			}
		}
	}
}


void PhotonMap::addElement
(
	const Light&        power,
	const Vector3f&     position,
	const UnitVector3f& direction,
	const dword         id
)
{
	/// check photons capacity wont overflow
	if( getRemainingSpace() > 0 )
	{
		const hxa7241::graphics::CieXyz powerXyz( power.getXYZ() );

		/// add photon
		photons_m.push_back( Photon( powerXyz, position, direction, id ) );

		/// update photons stats
		minPosition_m.clampToMaxOf( position );
		maxPosition_m.clampToMinOf( position );
	}
}


void PhotonMap::scale
(
	const float scaling
)
{
	for( dword i = photons_m.size();  i-- > 0; )
	{
		photons_m[i].scalePower( scaling );
	}
}




/// queries ----------------------------------------------------------------------------------------
dword PhotonMap::getCapacity() const
{
	return photons_m.capacity();
}


dword PhotonMap::getRemainingSpace() const
{
	return photons_m.capacity() - photons_m.size();
}


const vector<LightStream>* PhotonMap::getIllumination
(
	const Vector3f& position
) const
{
	const vector<LightStream>* pIllumination = 0;
	{
		dword visitCount  = 0;
		dword shrinkCount = 0;
		const bool isEstimated = getEstimateIlluminations( position, pOctree_m.get(), octreeMaxDepth_m, octreeElementCount_m, pBuffers_m,
		                                                   visitCount, shrinkCount );
		//const bool isEstimated = getEstimateIlluminationsTest( position, photons_m, pBuffers_m );
		if( isEstimated & (pBuffers_m != 0) )
		{
			pIllumination = &(pBuffers_m->getIlluminationBuffer());

			/// update measurements
			++totalEstimationCount_m;
//			totalVisitCount_m  += visitCount;
//			totalShrinkCount_m += shrinkCount;
//			maxVisitCount_m  = maxVisitCount_m  >= visitCount  ? maxVisitCount_m  : visitCount;
//			maxShrinkCount_m = maxShrinkCount_m >= shrinkCount ? maxShrinkCount_m : shrinkCount;
		}
	}

	return pIllumination;
}




/// implementation ---------------------------------------------------------------------------------
bool PhotonMap::getEstimateIlluminationsTest
(
	const Vector3f&       center,
	const vector<Photon>& photons,
	PhotonMapBuffers*     pBuffers
)
{
	const bool isEstimationPossible = (pBuffers != 0) && (pBuffers->getCapacity() > 0);

	if( isEstimationPossible )
	{
		/// make sortable index of photons
		vector<SelectedItem>& photonIndex = pBuffers->getSelectedBuffer();
		photonIndex.resize( photons.size() );
		for( dword i = photonIndex.size();  i-- > 0; )
		{
			const Photon* pp = &(photons[i]);

			const Vector3f& photonPosition = pp->getPosition();
			const float dx = photonPosition.getX() - center.getX();
			const float dy = photonPosition.getY() - center.getY();
			const float dz = photonPosition.getZ() - center.getZ();
			const float distance2 = (dx * dx) + (dy * dy) + (dz * dz);

			//const Vector3f dif( pp->getPosition() - center );
			//const float distance2 = dif.dot( dif );
			//const float distance = center.distance( pp->getPosition() );

			photonIndex[i] = SelectedItem( distance2, pp );
		}


		/// select nearest photons
		{
			const dword selectionLength = pBuffers->getCapacity();
			std::partial_sort( photonIndex.begin(), photonIndex.begin() + selectionLength, photonIndex.end(), SelectedComparer() );
			//std::sort( photonIndex.begin(), photonIndex.end(), SelectedComparer() );

			const float selectionFurthest = ::sqrt( photonIndex[selectionLength - 1].first );
			const float selectionRadius   = selectionFurthest + (selectionFurthest / float(0x00100000));

			pBuffers->resize( selectionLength );
			for( dword i = 0;  i < selectionLength;  ++i )
			{
				const SelectedItem& selected = photonIndex[i];
				const Photon* pPhoton = selected.second;
				/// pPhoton never zero, because index making
				{
					/// extract photon data
					CieXyz       power;
					Vector3f     position;
					UnitVector3f direction;
					dword        id;
					pPhoton->get( power, position, direction, id );

					/// scale power
					const float weighting = hxa7241::graphics::pavicicFilter( selectionRadius, ::sqrt( selected.first ) );
					power *= weighting;
					//power /= 3.1415926535897932384626433832795f * selectionRadius * selectionRadius;

					/// copy to illuminations
					pBuffers->setElement( i, LightStream( direction, Light(power) ), selected.first, id );
				}
			}
		}
	}

	return isEstimationPossible;
}


bool PhotonMap::getEstimateIlluminations
(
	const Vector3f&       center,
	const Octree<Photon>* pOctree,
	const dword           octreeMaxDepth,
	const dword           octreeElementCount,
	PhotonMapBuffers*     pBuffers,
	dword&                visitCount,
	dword&                shrinkCount
)
{
	/// check if octree is not empty, and neighbour count is larger than zero
		/// fill illumination buffer
			/// prepare to loop
				/// set shell radiuss
					/// start small...
			///	loop through expanding shells
				/// zero length of selected photons buffer (distance, photon ptr)
				/// call octree to select
				/// add selected photons to illumination buffer
					/// if all selected wont fit
						/// sort selected photons
					/// add all selected photons to illuminations, or until illuminations full
						/// convert photon to illumination
				/// increase shell radiuss
					///
		/// scale and weight illumination buffer
		/// get bounced light
	/// else
		/// set light to zero


	/// zero counters
	visitCount  = 0;
	shrinkCount = 0;

	/// check theres an octree of non zero size with items in it and an estimation buffer with space available
	const bool isOctreeOk  = (pOctree  != 0) && (!pOctree->isEmpty() & (pOctree->getSize() > 0));
	const bool isBuffersOk = (pBuffers != 0) && (pBuffers->getCapacity() > 0);
	const bool isEstimationPossible = isOctreeOk & isBuffersOk;

	if( isEstimationPossible )
	{
		pBuffers->resize( 0 );

		///// calculate shell start based on average density of octree
		static const float EXTRA_BIT               = 1.125f;
		static const float DOUBLE_VOLUME           = 1.259921f;
		//static const float ONE_OVER_FOUR_THIRDS_PI = 3.0 / (4.0 * 3.1415926535897932384626433832795);
		//float startingSphereRadius;
		//{
		//	const float octreeSize   = pOctree->getSize();
		//	const float octreeVolume = octreeSize * octreeSize * octreeSize;
		//	const float octreeDensity        = float(octreeElementCount) / octreeVolume;
		//	const float startingSphereVolume = (float(pBuffers->getCapacity()) * EXTRA_BIT) / octreeDensity;
		//	startingSphereRadius             = float(::pow( (startingSphereVolume * ONE_OVER_FOUR_THIRDS_PI), 1.0f/3.0f ));
		//}

		/// calculate shell start based on maximum density in octree
		static const float ONE_OVER_FOUR_THIRDS_PI = 3.0f / (4.0f * 3.1415926535897932384626433832795f);
		float startingSphereRadius;
		float oneOverRoughMaxDensity;
		{
			const float maxDensityCellWidth  = pOctree->getSize() / float(1 << (octreeMaxDepth - 1));
			const float roughMaxDensity      = (float(pOctree->getMaxElementCountPerCell()) * 0.5f) / (maxDensityCellWidth * maxDensityCellWidth * maxDensityCellWidth);
			oneOverRoughMaxDensity           = 1.0f / roughMaxDensity;
			const float startingSphereVolume = float(pBuffers->getCapacity()) * oneOverRoughMaxDensity * 2.0f;
			startingSphereRadius             = ::pow( (startingSphereVolume * ONE_OVER_FOUR_THIRDS_PI), 1.0f/3.0f );
//hxa7241::utility::Logger() << "PhotonMap::getEstimateIlluminations  maxDensityCellWidth=" << maxDensityCellWidth << " roughMaxDensity=" << roughMaxDensity << " startingSphereVolume=" << startingSphereVolume << " startingSphereRadius=" << startingSphereRadius;   /// DEBUG ///
//hxa7241::utility::Logger() << "PhotonMap::getEstimateIlluminations  starting sphere count = " << ((1.0f / ONE_OVER_FOUR_THIRDS_PI) * startingSphereRadius * startingSphereRadius * startingSphereRadius * roughMaxDensity);   /// DEBUG ///
		}

		/// set starting visit conditions
		float radiusInner = 0.0f;
		float radiusOuter = startingSphereRadius;

		/// accumulate photons in increasing size sphere shells
		/// (a 'sphere shell' being: the 3d equivalent of a letter 'o', a spherical region of just the crust and discluding the core)
		while( !pBuffers->isFull() & (pBuffers->getSize() < octreeElementCount) )
		{

			/// visit octree to select items
			bool isSelectedOverflow;
			{
				vector<SelectedItem>& selectedBuffer = pBuffers->getSelectedBuffer();
				selectedBuffer.resize( 0 );
				OctreeVisitorPhoton octreeVisitor( &selectedBuffer, center, radiusOuter, radiusInner );

				pOctree->visitItems( octreeVisitor );
				++visitCount;

				isSelectedOverflow = udword(octreeVisitor.getVisitCount()) > selectedBuffer.capacity();
//hxa7241::utility::Logger() << "PhotonMap::getEstimateIlluminations  radiusInner=" << radiusInner << "  radiusOuter=" << radiusOuter << "    isSelectedOverflow=" << isSelectedOverflow << "  " << octreeVisitor.getVisitCount();   /// DEBUG ///
			}

			/// check that selected buffer didnt overflow
			if( !isSelectedOverflow )
			{
				/// add to illuminations
				addSelectedPhotonsToIlluminations( pBuffers );

				/// expand shell by scaling its volume in proportion to remaining selections needed
				{
					radiusInner = radiusOuter;

					if( pBuffers->getSize() > 0 )
					{
						radiusOuter *= float(::pow( (float(pBuffers->getCapacity()) / float(pBuffers->getSize())) * EXTRA_BIT, 1.0f/3.0f ));
					}
					else
					{
						radiusOuter *= DOUBLE_VOLUME;
					}

					///// v = (4/3 * PI * ro^3) - (4/3 * PI * ri^3)
					///// v = (4/3 * PI) * (ro^3 - ri^3)
					///// v * (3 / (4 * PI)) = ro^3 - ri^3
					///// (v * (3 / (4 * PI))) + ri^3 = ro^3
					///// ro = cuberoot( (v * (3 / (4 * PI))) + ri^3 )
					//const float nextShellVolume = pBuffers->getSpare() * oneOverRoughMaxDensity;
					//const float radiusInner3 = radiusInner * radiusInner * radiusInner;
					//radiusOuter = float(::pow( ((nextShellVolume * ONE_OVER_FOUR_THIRDS_PI) + radiusInner3), 1.0f/3.0f ));
				}
			}
			else
			{
				/// shrink shell by halving its volume
				/// v = (4/3 * PI * ro^3) - (4/3 * PI * ri^3)
				/// ((4/3 * PI * ro^3) - (4/3 * PI * ri^3)) /2 = (4/3 * PI * rm^3) - (4/3 * PI * ri^3)
				/// (2/3 * PI) * (ro^3 - ri^3) = (4/3 * PI) * (rm^3 - ri^3)
				/// 1/2 * (ro^3 - ri^3) = rm^3 - ri^3
				/// ((ro^3 - ri^3) / 2) + ri^3 = rm^3
				/// rm = cuberoot( (ro^3 + ri^3) / 2 )
				const float radiusInner3 = radiusInner * radiusInner * radiusInner;
				const float radiusOuter3 = radiusOuter * radiusOuter * radiusOuter;
				radiusOuter = float(::pow( (radiusInner3 + radiusOuter3) * 0.5f, 1.0f/3.0f ));

				++shrinkCount;
			}
		}
//hxa7241::utility::Logger(true) << "PhotonMap::getEstimateIlluminations  visitCount=" << visitCount;   /// DEBUG ///

		filterIlluminations( pBuffers->getDistance2Buffer(), pBuffers->getIlluminationBuffer() );
	}

	return isEstimationPossible;
}


void PhotonMap::addSelectedPhotonsToIlluminations
(
	PhotonMapBuffers* pBuffers
)
{
	vector<SelectedItem>& selectedBuffer = pBuffers->getSelectedBuffer();

	/// determine if selecteds will fill illuminations
	const dword selectedLength     = selectedBuffer.size();
	const dword illuminationsSpare = pBuffers->getSpare();
	const bool  isAboutToBeFilled  = selectedLength >= illuminationsSpare;
//if( selectedLength > 0 ) hxa7241::utility::Logger() << "\tPhotonMap::addSelectedPhotonsToIlluminations  selectedLength=" << selectedLength << " illuminationsSpare=" << illuminationsSpare << " isAboutToBeFilled=" << isAboutToBeFilled;   /// DEBUG ///

	/// sort last selection, since some wont be included - and they should be the furthest ones,
	/// and the last distance buffer element must be the furthest
	if( isAboutToBeFilled )
	{
		std::partial_sort( selectedBuffer.begin(), selectedBuffer.begin() + illuminationsSpare, selectedBuffer.end(), SelectedComparer() );
		//std::sort( selectedBuffer.begin(), selectedBuffer.end(), SelectedComparer() );
		//std::stable_sort( selectedBuffer.begin(), selectedBuffer.end(), SelectedComparer() );
	}

	/// convert selected items
	for( dword i = 0, count = isAboutToBeFilled ? illuminationsSpare : selectedLength;  i < count;  ++i )
	{
		const SelectedItem& selected = selectedBuffer[i];
		const Photon* pPhoton = selected.second;
		if( pPhoton != 0 )
		{
			/// extract photon data
			CieXyz       power;
			Vector3f     position;
			UnitVector3f direction;
			dword        id;
			pPhoton->get( power, position, direction, id );

			/// append to illuminations
			pBuffers->pushBack( LightStream( direction, Light(power) ), selected.first, id );
//hxa7241::utility::Logger() << "\tPhotonMap::addSelectedPhotonsToIlluminations  " << i << " LightStream=" << LightStream( direction, Light(power) );   /// DEBUG ///
		}
	}
}


void PhotonMap::filterIlluminations
(
	const vector<float>& distance2Buffer,
	vector<LightStream>& illuminationBuffer
)
{
	if( !distance2Buffer.empty() )
	{
		const float estimateRadius2 = distance2Buffer.back() ;//+ (distance2Buffer.back() / float(0x00100000));

		for( int i = illuminationBuffer.size();  i-- > 0; )
		{
			const float weighting = hxa7241::graphics::pavicicFilter2( estimateRadius2, distance2Buffer[i] );
			//const float weighting = hxa7241::graphics::cookFilter2( estimateRadius2, distance2Buffer[i] );
			//const float weighting = 1.0f / (3.1415926535897932384626433832795 * estimateRadius2);

			illuminationBuffer[i].light_m *= weighting;
		}
	}
}








/// PhotonMapBuffers ///////////////////////////////////////////////////////////////////////////////

/// standard object services -----------------------------------------------------------------------
PhotonMapBuffers::PhotonMapBuffers()
 :	illuminationBuffer_m(),
	distance2Buffer_m(),
	idBuffer_m(),
	selectedBuffer_m()
{
}


PhotonMapBuffers::~PhotonMapBuffers()
{
}




/// commands ---------------------------------------------------------------------------------------
void PhotonMapBuffers::reserve
(
	dword amount
)
{
	if( amount < 0 )
	{
		amount = 0;
	}

	if( illuminationBuffer_m.capacity() < udword(amount) )
	{
		illuminationBuffer_m.reserve( udword(amount) );
	}

	if( distance2Buffer_m.capacity() < udword(amount) )
	{
		distance2Buffer_m.reserve( udword(amount) );
	}

	if( idBuffer_m.capacity() < udword(amount) )
	{
		idBuffer_m.reserve( udword(amount) );
	}


	if( selectedBuffer_m.capacity() < udword(amount * 2) )
	{
		selectedBuffer_m.reserve( udword(amount * 2) );
	}
}


void PhotonMapBuffers::resize
(
	dword amount
)
{
	if( amount < 0 )
	{
		amount = 0;
	}

	illuminationBuffer_m.resize( udword(amount) );
	distance2Buffer_m.resize( udword(amount) );
	idBuffer_m.resize( udword(amount) );
}


void PhotonMapBuffers::pushBack
(
	const LightStream& lightstream,
	const float        distance2,
	const dword        id
)
{
	illuminationBuffer_m.push_back( lightstream );
	distance2Buffer_m.push_back( distance2 );
	idBuffer_m.push_back( id );
}


void PhotonMapBuffers::setElement
(
	const dword        index,
	const LightStream& lightstream,
	const float        distance2,
	const dword        id
)
{
	illuminationBuffer_m[index] = lightstream;
	distance2Buffer_m[index]    = distance2;
	idBuffer_m[index]           = id;
}




/// queries ----------------------------------------------------------------------------------------
dword PhotonMapBuffers::getSpare() const
{
	return illuminationBuffer_m.capacity() - illuminationBuffer_m.size();
}








/*void OctreeRoot::visitItemsNearestToPoint
(
	const Vector3f& position,
	const dword     numberOfItems,
	OctreeVisitor&  visitor
) const
{
	///	input query point
	///
	///	construct sorted set of items
	///		has max possible size, disallows duplicates, always sorted by distance from query point
	///
	///	set sphere shell inner to zero
	///	set sphere shell outer based on octree density (item count / root cell volume)
	///		outer radius = cuberoot( (number of nearest items / density) / (4/3 * pi) ) * a bit extra
	///
	///	loop while item set not full
	///		visit nodes intersecting sphere shell
	///			add leaf node items intersecting sphere shell to item set
	///		expand sphere shell
	///			set sphere shell inner to outer
	///			set sphere shell outer based on density of last harvest
	///
	///	visit items in item set


	if( pRootCell_m != 0 )
	{
		if( numberOfItems > 0 )
		{
			/// only act if point inside octree
			if( data_m.isOverlappingBox( position, Vector3f::ZERO() ) != OctreeBound::OUTSIDE )
			{
				ItemSet items;

				/// find items
				{
					/// set initial shell size
					static const float THREE_OVER_FOUR_PI = 3.0 / (4.0 * 3.1415926535897932384626433832795);
					static const float EXTRA_BIT          = 1.1f;
					static const float DOUBLE_VOLUME      = 1.26f;
					const float octreeSize   = data_m.getBound().getSize();
					const float octreeVolume = (octreeSize * octreeSize * octreeSize);
					const dword octreeItems  = data_m.getTotalItemCount();

					if( (octreeItems > 0) & (octreeVolume > 0.0f) )
					{
						const float octreeDensity = float(octreeItems) / octreeVolume;

						float radiusInner = 0.0f;
						float radiusOuter = float(::pow( THREE_OVER_FOUR_PI * (float(numberOfItems) / octreeDensity), 1.0/3.0 )) * EXTRA_BIT;

						/// loop while item set not full
						while( items.size() < udword(numberOfItems) )
						{
							pRootCell_m->visitItemsOverlappingShell( data_m, position, radiusOuter, radiusInner, false, visitor, items );

							/// expand shell (can only shrink if item set is full)
							radiusInner = radiusOuter;
							if( items.size() > 0 )
							{
								radiusOuter *= float(::pow( float(numberOfItems) / float(items.size()), 1.0/3.0 )) * EXTRA_BIT;
							}
							else
							{
								radiusOuter *= DOUBLE_VOLUME;
							}
						}
					}
				}

				/// visit items
				{
					ItemSet::ConstIterator it( items.getBegin() );
					ItemSet::ConstIterator end( items.getEnd() );
					for( dword i = 0;  (it != end) & (i < numberOfItems);  ++it, ++i )
					{
						visitor.visitItemT( it->second );
					}
				}
			}
		}
	}
}*/






/// testing ----------------------------------------------------------------------------------------
#ifndef ENABLE_TEST_CODE
bool PhotonMap::test
(
	const bool isLogging
)
{
	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  [disabled]";

	return false;
}


#else
#include "RandomGood.h"
#include <time>
#include <set>

bool PhotonMap::test
(
	const bool isLogging
)
{
	bool isSucceeded = true;

	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  === general test ===";


	static const dword    PHOTON_TESTS   = 10;
	static const dword    ESTIMATE_TESTS = 100;
	static const dword    CAPACITY   = 10000;
	static const dword    ESTIMATION = ESTIMATION_SIZE;//CAPACITY / ESTIMATION_PROPORTION;
	static const float    SIZE       = 3.7f;
	static const Vector3f POSITION( -1.9f, -0.2f, 10.0f );

	const dword seed = dword(std::time(0));
	hxa7241::utility::RandomGood rand( seed );
	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  seed " << seed;


	dword totalEstimationCount = 0;
	dword totalVisitCount = 0;
	dword totalShrinkCount = 0;
	dword maxVisitCount = 0;
	dword maxShrinkCount = 0;


	for( dword dd = PHOTON_TESTS;  (dd-- > 0) & isSucceeded; )
	{
		if(isLogging) hxa7241::utility::Logger() << "\tPhotonMap::test  /// new photons";

		PhotonMapBuffers pmb;
		PhotonMap pm( &pmb, CAPACITY, true );


		/// check element insertion
		{
			/// add random elements, > capacity
			{
				for( dword c = 0, end = CAPACITY + 16;  c < end;  ++c )
				{
					const float x = rand.getFloat();
					rand.next();
					const float y = rand.getFloat();
					rand.next();
					const float z = rand.getFloat();
					rand.next();

					pm.addElement( Light::ONE(), (Vector3f( x, y, z ) *= SIZE) += POSITION, UnitVector3f::X(), c );
				}
			}

			/// check photons_m size and capacity
			/// check getCapacity and getRemainingSpace
			/// check maxPosition_m, minPosition_m
			const bool isCapacityOk    = pm.photons_m.capacity() == CAPACITY;
			const bool isAddElementsOk = pm.getRemainingSpace() == 0;

			isSucceeded &= isCapacityOk;
			isSucceeded &= isAddElementsOk;
			if( isLogging )
			{
				if(!isCapacityOk)    hxa7241::utility::Logger() << "PhotonMap::test  *** capacity failure  " << CAPACITY << "  " << pm.photons_m.capacity();
				if(!isAddElementsOk) hxa7241::utility::Logger() << "PhotonMap::test  *** addElements failure  " << pm.getRemainingSpace();
			}

			if( isLogging )
			//if( isCapacityOk & isAddElementsOk & isLogging )
			{
				hxa7241::utility::Logger() << "PhotonMap::test  capacity=" << CAPACITY << " estimation=" << ESTIMATION;
				hxa7241::utility::Logger() << "PhotonMap::test  pm.photons_m.capacity()=" << pm.photons_m.capacity() << " photons_m.size=" << pm.photons_m.size() << " getRemainingSpace=" << pm.getRemainingSpace();
				hxa7241::utility::Logger() << "PhotonMap::test  minPosition_m=" << pm.minPosition_m << " maxPosition_m=" << pm.maxPosition_m;
			}
		}


		/*/// check scale
		{
			Light preSum;
			{
				for( dword i = 0, end = pm.photons_m.size();  i < end;  ++i )
				{
					/// extract photon data
					CieXyz       power;
					Vector3f     position;
					UnitVector3f direction;
					dword        id;
					pm.photons_m[i].get( power, position, direction, id );

					Light p( power );
					preSum += p;
				}
			}

			const float scale = 2.3f;
			pm.scale( scale );

			Light postSum;
			{
				for( dword i = 0, end = pm.photons_m.size();  i < end;  ++i )
				{
					/// extract photon data
					CieXyz       power;
					Vector3f     position;
					UnitVector3f direction;
					dword        id;
					pm.photons_m[i].get( power, position, direction, id );

					Light p( power );
					postSum += p;
				}
			}

			if(isLogging) hxa7241::utility::Logger( true ) << "PhotonMap::test  preSum=" << preSum.average() << " scale=" << scale << " postSum=" << postSum.average() << " preSum*scale=" << (preSum.average() * scale) << "  dif=" << ((postSum.average() - (preSum.average() * scale)) / postSum.average());
		}*/


		for( dword cc = ESTIMATE_TESTS;  (cc-- > 0) & isSucceeded; )
		{
			//if(isLogging) hxa7241::utility::Logger() << "\tPhotonMap::test  /// new estimation";

			dword estimationSize = ESTIMATION;

			const float x = rand.getFloat();
			rand.next();
			const float y = rand.getFloat();
			rand.next();
			const float z = rand.getFloat();
			rand.next();
			const Vector3f center( (Vector3f( x, y, z ) *= SIZE) += POSITION );

			//if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  center=" << center << " estimationSize=" << estimationSize;

			/// exhaustive
			PhotonMapBuffers exb;
			{
				vector<SelectedItem>& photonIndex = exb.getSelectedBuffer();
				photonIndex.resize( pm.photons_m.size() );
				for( dword i = photonIndex.size();  i-- > 0; )
				{
					const Photon* pp = &(pm.photons_m[i]);

					const Vector3f v( pp->getPosition() - center );
					const float d2 = v.dot( v );

					photonIndex[i] = SelectedItem( d2, pp );
				}
				std::sort( photonIndex.begin(), photonIndex.end(), SelectedComparer() );

				exb.reserve( estimationSize );
				for( dword i = 0;  i < estimationSize;  ++i )
				{
					const SelectedItem& selected = photonIndex[i];
					const Photon* pPhoton = selected.second;
					if( pPhoton != 0 )
					{
						/// extract photon data
						CieXyz       power;
						Vector3f     position;
						UnitVector3f direction;
						dword        id;
						pPhoton->get( power, position, direction, id );

						/// append to illuminations
						exb.pushBack( LightStream( direction, Light(power) ), selected.first, id );
					}
				}

				//pm.filterIlluminations( exb.getDistance2Buffer(), exb.getIlluminationBuffer() );
			}

			/// octree
			{
				pm.switchToIlluminatingMode();
				dword visitCount;
				dword shrinkCount;
				const bool isEstimated = getEstimateIlluminations( center, pm.pOctree_m.get(), pm.octreeMaxDepth_m, pm.octreeElementCount_m, pm.pBuffers_m,
				                                                   visitCount, shrinkCount );
				if( isEstimated )
				{
					++totalEstimationCount;
					totalVisitCount += visitCount;
					totalShrinkCount  += shrinkCount;
					maxVisitCount  = maxVisitCount  >= visitCount  ? maxVisitCount  : visitCount;
					maxShrinkCount = maxShrinkCount >= shrinkCount ? maxShrinkCount : shrinkCount;
				}
			}


			/// compare
			{
				/*if( isLogging )
				{
					const dword size0 = exb.getSize();
					const dword size1 = pmb.getSize();
					const dword size  = size0 < size1 ? size0 : size1;
					for( dword i = 0;  i < size;  ++i )
					{
						hxa7241::utility::Logger() << "\t" << i << "    " << exb.getDistance2Buffer()[i] << "  " << exb.getIdBuffer()[i] << "    " << pmb.getDistance2Buffer()[i] << "  " << pmb.getIdBuffer()[i];
					}
				}*/


				const bool isLengthsSame = exb.getSize() == pmb.getSize();

				isSucceeded &= isLengthsSame;
				if(isLogging) if(!isLengthsSame) hxa7241::utility::Logger() << "PhotonMap::test  *** estimation sizes unmatched  " << exb.getSize() << "  " << pmb.getSize();

				if( isLengthsSame & (exb.getSize() > 0) )
				{
					const bool isLastSame = exb.getDistance2Buffer().back() == pmb.getDistance2Buffer().back();

					isSucceeded &= isLastSame;
					if(isLogging) if(!isLastSame) hxa7241::utility::Logger(true) << "PhotonMap::test  *** last unmatched  " << (exb.getSize()-1) << "    " << exb.getDistance2Buffer().back() << "  " << exb.getIdBuffer().back() << "    " << pmb.getDistance2Buffer().back() << "  " << pmb.getIdBuffer().back() << "    " << (exb.getDistance2Buffer().back()  - pmb.getDistance2Buffer().back());

					{
						bool isOrdered = true;
						for( dword i = exb.getSize() - 1;  (i-- > 0) & isOrdered; )
						{
							isOrdered =  exb.getDistance2Buffer()[i] <= exb.getDistance2Buffer()[i+1];
						}

						isSucceeded &= isOrdered;
						if(isLogging) if(!isOrdered) hxa7241::utility::Logger() << "PhotonMap::test  *** unordered  ";
					}

					{
						std::set<dword> idset;

						const float lastDist2 = exb.getDistance2Buffer().back();
						for( dword i = exb.getSize();  i-- > 0; )
						{
							if( exb.getDistance2Buffer()[i] != lastDist2 )
							{
								idset.insert( exb.getIdBuffer()[i] );
								//if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  set ins: " << exb.getIdBuffer()[i];
							}
						}

						bool isUnmatching0 = false;
						for( dword i = exb.getSize();  (i-- > 0) & !isUnmatching0; )
						{
							if( pmb.getDistance2Buffer()[i] != lastDist2 )
							{
								isUnmatching0 = idset.erase( pmb.getIdBuffer()[i] ) != 1;
								//if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  set del: " << pmb.getIdBuffer()[i] << "  " << (isUnmatching0 ? "not found" : "");
							}
						}

						bool isUnmatching1 = !idset.empty();

						isSucceeded &= !isUnmatching0;
						isSucceeded &= !isUnmatching1;
						if(isLogging) if(isUnmatching0 | isUnmatching1) hxa7241::utility::Logger() << "PhotonMap::test  *** estimate collection unmatching";
					}
				}
			}
		}

		if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  " << "pmb.getSelectedBuffer().capacity()=" << pmb.getSelectedBuffer().capacity() << "  pmb.getCapacity()=" << pmb.getCapacity();
	}


	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  " << "totalEstimationCount=" << totalEstimationCount << "  totalVisitCount=" << totalVisitCount << "  totalShrinkCount=" << totalShrinkCount << "  totalExpandCount=" << (totalVisitCount - totalShrinkCount);
	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  " << "meanVisitCount=" << (float(totalVisitCount) / float(totalEstimationCount)) << "  meanShrinkCount=" << (float(totalShrinkCount) / float(totalEstimationCount)) << "  meanExpandCount=" << (float(totalVisitCount - totalShrinkCount) / float(totalEstimationCount));
	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  " << "maxVisitCount=" << maxVisitCount << "  maxShrinkCount=" << maxShrinkCount;


	if(isLogging) hxa7241::utility::Logger() << "PhotonMap::test  " << (isSucceeded ? "--- PhotonMap test succeeded" : "*** PhotonMap test failed");

	return isSucceeded;
}
#endif
