/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef RasterizerProjective_h
#define RasterizerProjective_h


#include "Primitives.h"
#include "Camera.h"
#include "RasterizationVertex.h"
#include "RasterizationTriangleIm.h"
#include "ProjectableObjectRef.h"
#include "Mesh.h"
#include "VectorEx.h"
#include <vector>

#include "perceptuum2_imaging.h"




#include "perceptuum2_transport_projection.h"
namespace perceptuum2 { namespace transport { namespace projection {
	using namespace hxa7241::primitives;
	using perceptuum2::transport::Camera;
	using perceptuum2::transport::Projection;
	using perceptuum2::transport::EyePath;
	using perceptuum2::transport::SamplingControl;
	using perceptuum2::transport::SpacetimeIndex;
	using perceptuum2::transport::Illumination;
	using perceptuum2::imaging::ImageAccumulation;
	using perceptuum2::modelling::ProjectableObjectRef;
	using perceptuum2::modelling::Mesh;
	using perceptuum2::modelling::ObjectRenderable;
	using perceptuum2::modelling::TriangleImplicit;
	using hxa7241::graphics::ClipFlags;
	using hxa7241::graphics::Vector3f;
	using hxa7241::utility::Vector;
	using std::vector;


/**
 *
 * references:<br/><br/>
 *
 * <cite>ch13, ch17, ch18
 * 'jim blinns corner, a trip down the graphics pipeline' - blinn
 * (morgan kaufmann 1996)</cite><br/><br/>
 *
 * <cite>'how to write a (software) 3d polygon pipeline' - bloom
 * (www.cbloom.com 2000)</cite><br/><br/>
 *
 * <cite>'what a 3d engine should be like' - bloom
 * (www.cbloom.com 2000)</cite>
 *
 * @implementation
 * first: for each object in the scene, get a pointer to its mesh, a transform to
 * take it from model to world space, and clipflags summarising its position.<br/>
 * then: for each object,
 * first: transform its vertexs into world and screen space, and set clipflags for
 * each vertex.
 * second: select its visible triangles using clipflags and facing orientation,
 * including adding triangles made by partly clipping others.
 * third:
 */
class RasterizerProjective
{
/// standard object services -----------------------------------------------------------------------
public:
	         RasterizerProjective( const Projection& projection,
	                               dword             imageWidth,
	                               dword             imageHeight,
	                               dword             eyePathLengthLimit,
	                               bool              isLogging );

	virtual ~RasterizerProjective();
private:
	         RasterizerProjective( const RasterizerProjective& );
	RasterizerProjective& operator=( const RasterizerProjective& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  rasterizeFrame( const SamplingControl&,
	                              const SpacetimeIndex&,
	                              const Illumination&,
	                              ImageAccumulation& );


/// queries ----------------------------------------------------------------------------------------
public:


/// implementation ---------------------------------------------------------------------------------
protected:
	static  void  transformVertexs( const Projector&             projector,
	                                const ClipFlags              objectClipflags,
	                                const Mesh&                  mesh,
	                                Vector<RasterizationVertex>& rasterVertexs );
	static  void  selectTriangles( const Projector&                 projector,
	                               const vector<TriangleImplicit>&  triangles,
	                               Vector<RasterizationTriangleIm>& rasterTriangles,
	                               Vector<RasterizationVertex>&     rasterVertexs );
	static  void  rasterizeMesh( const Vector<RasterizationVertex>&     vertexs,
	                             const Vector<RasterizationTriangleIm>& triangles,
	                             const ObjectRenderable&                object,
	                             const Vector3f&                        eyePosition,
	                             const SpacetimeIndex&                  sceneIndex,
	                             const Illumination&                    illumination,
	                             ImageAccumulation&                     accumulation );


/// fields -----------------------------------------------------------------------------------------
private:
	Camera camera_m;

	vector<ProjectableObjectRef>    objectBuffer_m;
	Mesh                            meshBuffer_m;
	Vector<RasterizationVertex>     rasterVertexBuffer_m;
	Vector<RasterizationTriangleIm> rasterTriangleBuffer_m;

	bool isLogging_m;
};


}}}




#endif//RasterizerProjective_h
